use super::*;

impl<A: AtomicModeParse> Atomic for A {
    fn a_deserialize(inlining: impl Inlining) -> AParseResult<Self> {
        Self::ma_deserialize(inlining).map(Self::seal)
    }

    fn a_extend(self, tail: &[u8]) -> AParseResult<Self> {
        Self::xseal(Self::ma_extend(Self::prepare(self), tail))
    }
}

/// [`Atomic`] equivalent of [`ModeResult`].
pub type AModeResultM<A, I> = ModeResultP<A, A, AParseError<A>, I>;

/// [`Atomic`] equivalent of [`ExtensionResult`].
pub type AExtensionResultM<A> = ExtensionResultP<A, A, AParseError<A>>;

/// [`Atomic`] equivalent of [`ExtensionSource`].
pub type AExtensionSourceM<A> = ExtensionSourceP<A, A>;

/// A more generic version of [`Atomic`].
pub trait AtomicModeParse: AtomicBase + ParseMode {
    /// A more generic version of [`Atomic::a_deserialize`].
    fn ma_deserialize<I: Inlining>(inlining: I) -> AModeResultM<Self, I>;

    /// A more generic version of [`Atomic::a_extend`].
    fn ma_extend(atomic: AExtensionSourceM<Self>, tail: &[u8]) -> AExtensionResultM<Self>;
}

/// External implementation of [`AtomicModeProxy`].
pub trait AtomicModeProxy {
    /// Associated [`AtomicModeParse`].
    type A: AtomicBase + ParseMode;

    /// External implementation of [`AtomicModeParse::ma_deserialize`].
    fn pma_deserialize<I: Inlining>(inlining: I) -> AModeResultM<Self::A, I>;

    /// External implementation of [`AtomicModeParse::ma_extend`].
    fn pma_extend(atomic: AExtensionSourceM<Self::A>, tail: &[u8]) -> AExtensionResultM<Self::A>;
}

impl<A: AtomicBase + WithParseMode> AtomicModeParse for A
where
    <A as WithParseMode>::WithMode: AtomicModeProxy<A = A>,
{
    fn ma_deserialize<I: Inlining>(inlining: I) -> AModeResultM<Self, I> {
        <<A as WithParseMode>::WithMode as AtomicModeProxy>::pma_deserialize(inlining)
    }

    fn ma_extend(atomic: AExtensionSourceM<Self>, tail: &[u8]) -> AExtensionResultM<Self> {
        <<A as WithParseMode>::WithMode as AtomicModeProxy>::pma_extend(atomic, tail)
    }
}

/// For auto-deriving [`RegularAtomic`] from concrete implementations.
pub trait CRegularAtomic: AtomicBase + ImplMode<Mode = RegularMode> {
    /// Concrete implementation of [`RegularAtomic::ra_deserialize`].
    fn cra_deserialize(inlining: impl Inlining) -> AParseResult<Self>;
    /// Concrete implementation of [`RegularAtomic::ra_extend`].
    fn cra_extend(self, tail: &[u8]) -> AParseResult<Self>;
}

/// Mostly same as [`AtomicModeParse`] but requires [`Mode`] to be [`RegularMode`].
pub trait RegularAtomic: AtomicBase + ParseMode<Mode = RegularMode> {
    /// Same as [`AtomicModeParse::ma_deserialize`].
    fn ra_deserialize(inlining: impl Inlining) -> AParseResult<Self>;
    /// Same as [`AtomicModeParse::ma_extend`].
    fn ra_extend(self, tail: &[u8]) -> AParseResult<Self>;
}

impl<A: AtomicModeParse + ParseMode<Mode = RegularMode>> RegularAtomic for A {
    fn ra_deserialize(inlining: impl Inlining) -> AParseResult<Self> {
        A::ma_deserialize(inlining)
    }

    fn ra_extend(self, tail: &[u8]) -> AParseResult<Self> {
        A::ma_extend(self, tail)
    }
}

impl<A: CRegularAtomic> AtomicModeProxy for WithMode<A, RegularMode> {
    type A = A;

    fn pma_deserialize<I: Inlining>(inlining: I) -> AModeResultM<Self::A, I> {
        A::cra_deserialize(inlining)
    }

    fn pma_extend(atomic: AExtensionSourceM<Self::A>, tail: &[u8]) -> AExtensionResultM<Self::A> {
        atomic.cra_extend(tail)
    }
}
